# SPDX-FileCopyrightText: 2023 DeepLime <contact@deeplime.io>
# SPDX-License-Identifier: MIT

import os
from typing import Any, List, Optional

from ...base.decorator import check_type
from ...base.project import Project
from ..output_element import OutputElement


class PlotlyOutput(OutputElement):
    @check_type
    def __init__(
        self,
        key: str,
        value: str,
        label: Optional[str] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any
    ):
        """
        A Plotly chart with a label on top.

        !!! warning
            Required package: `plotly`.

        Args:
            key: ID of the element. It must be unique as it is the key used to store data in
                Project(), otherwise it will lead to conflicts at runtime in both execution and
                Streamlit modes. The key will be transformed into snake case and slugified to avoid
                any special character or whitespace. Note that an ID cannot start with `_`. Try to
                choose a key that is meaningful for your context (see
                [examples projects](https://github.com/deeplime-io/onecode/tree/main/examples)).
            value: Path to the output Plotly JSON file which must have a `.json` extension. Unless
                absolute, a path is relative to the `outputs` folder of the flow currently running.
            label: Typically to be used by Streamlit for display purpose only. If not defined, it
                will default to the `key`.
            tags: Optional meta-data information about the expected file. This information is only
                used when the JSON output attributes are written to the output manifest.
            **kwargs: Extra user meta-data to attach to the element. Argument names cannot overwrite
                existing attributes or methods name such as `streamlit`, `_value`, etc.

        Raises:
            ValueError: if the `key` is empty or starts with `_`.
            AttributeError: if one the `kwargs` conflicts with an existing attribute or method.

        !!! example
            ```py
            import plotly
            import plotly.graph_objects as go
            from onecode import plotly_output, Mode, Project

            Project().mode = Mode.EXECUTE
            Project().current_flow = 'test'

            plot_file = plotly_output(
                key="PlotlyOutput",
                value="/path/to/file.json",
                label="My PlotlyOutput",
                tags=['Graph']
            )

            fig = go.Figure(
                data=[go.Bar(x=[1, 2, 3], y=[1, 3, 2])],
                layout=go.Layout(
                    title=go.layout.Title(text="A Figure Specified By A Graph Object")
                )
            )
            plotly.io.write_json(fig, plot_file)
            print(plot_file)
            ```

            ```py title="Output"
            "/path/to/file.json"
            ```

        """
        super().__init__(
            key,
            value,
            label,
            tags=tags,
            **kwargs
        )

    @property
    def value(self) -> str:
        """
        Returns:
            The path to the output file: if path are not absolute, then it is considered relative
            to the data output folder. See [Best Practices With Data][best-practices-with-data]
            for more information.

        """
        return Project().get_output_path(self._value)

    @check_type
    def _validate(
        self,
        value: str
    ) -> None:
        """
        Raises:
            ValueError: if the file does not have a JSON extension `.json`.

        """
        _, ext = os.path.splitext(value)
        valid_ext = [
            '.json',
        ]

        if ext.lower() not in valid_ext:
            raise ValueError(
                f"[{self.key}] Invalid Plotly extension: {ext} (accepted: {', '.join(valid_ext)})"
            )

    @staticmethod
    def imports() -> List[str]:
        """
        Returns:
            Python import statements required by the Streamlit code.

        """
        return ["import plotly"]

    @staticmethod
    def streamlit() -> str:
        """
        Returns:
            The Streamlit code to show a figure corresponding to the output Plotly file.

        """
        return """
value = os.path.relpath(value)  # allows compat with Windows
if not os.path.exists(value) and not os.path.isfile(value):
    st.warning(f'Invalid file path: {{value}}')

else:
    fig = plotly.io.read_json(value)

    st.subheader(f'{label} - {os.path.basename(value)}')
    st.plotly_chart(fig)

"""
